==> Strawberry-PI-0.0.56.tar.gz: Strawberry-PI-0.0.56/src/LoRa/LoRa.py <==
import time

from .board_conf import Strawberry
from .constants import DEFAULT, PA_DAC, REG
from .my_types import IRQ, MODE, ModemConfig


class LoRa(object):
    # ID del dispositivo
    _device_id: bytes

    # Frequency of Transmission
    _frequency: float

    # Mode:
    #   - É la modalità di inizio, utilizzata per salvare la modalità quando si effettuano più operazioni
    #   -  É la modalità CURRENT
    _initial_mode: int | None = None
    _current_mode: int | None = None

    # To Unpack
    _unpack: bool = DEFAULT.UNPACK

    # -------------------------------------------------------------------------------------------- #
    @staticmethod
    def Setup(
        device_id: str,
        frequency: float = DEFAULT.FREQUENCY,
        tx_power: int = DEFAULT.TX_POWER,
        modem_config: ModemConfig = DEFAULT.MODEM_CONFIG,
        handler: object = None,
        unpack: bool = DEFAULT.UNPACK,
    ):
        # --- Device ID ---- #
        LoRa._device_id = LoRa.encode_devID(device_id)

        # --- SPI --- #
        Strawberry.SpiDev()  # Open Serial Peripheral
        Strawberry.Setup()  # Setup all GPIO
        Strawberry.AddEvents(
            callback_dio0=LoRa._handle_interrupt_dio0,
            callback_dio1=LoRa._handle_interrupt_dioX,
            callback_dio2=LoRa._handle_interrupt_dioX,
            callback_dio3=LoRa._handle_interrupt_dioX,
        )  # Add Events

        # --- Settings --- #
        LoRa.check_module()  # Check if module LoRa is connected
        LoRa.set_mode(MODE.SLEEP)  # For calibration
        # LoRa.reset_fifo()  # Reset Fifo buffer
        LoRa.set_modem_config(modem_config)  # Set Modem Config
        # self.set_preamble()  # Set Preamble
        LoRa.set_frequency(frequency)  # Frequency
        LoRa.set_tx_power(tx_power)  # Transmission Power
        LoRa.set_handler(handler, unpack)
        LoRa.set_mode(
            MODE.STANDBY
        )  # Module in Standby -> In questo modo sarà il Main a scegliere il da farsi

    # -------------------------------------------------------------------------------------------- #
    # --- Events --- #
    @staticmethod
    def on_receive(payload: list[int], snr: float, rssi: float) -> None:
        raise NotImplementedError(
            "Coglione ti sei dimenticato di implementare la funzione!"
        )

    @staticmethod
    def on_receive_unpacked(
        h_to: bytes, h_from: bytes, payload: bytes, snr: float, rssi: float
    ) -> None:
        raise NotImplementedError(
            "Coglione ti sei dimenticato di implementare la funzione!"
        )

    # -------------------------------------------------------------------------------------------- #
    # --- Module Function --- #
    @staticmethod
    def set_handler(callback: object, unpack: bool = False):
        if callback is None:
            # Imposto la callback per i messaggi che ricevo
            LoRa._unpack = unpack

            if unpack:
                LoRa.on_receive_unpacked = callback  # type: ignore
            else:
                LoRa.on_receive = callback  # type: ignore

    @staticmethod
    def check_module():
        # Controllo se il modulo è stato correttamente collegato al microcontrollore
        #     Scrivo nel registro e quando rileggo mi devo aspettare lo stesso risultato
        Strawberry.Write(REG.OP_MODE, MODE.SLEEP)
        time.sleep(0.1)

        if Strawberry.Read(REG.OP_MODE) != MODE.SLEEP:
            raise ConnectionError("LoRa initialization failed!")

    @staticmethod
    def reset_fifo():
        # Imposto la posizione di base del buffer FIFO di trasmissione ('FIFO TX') e di ricezione ('FIFO RX) a 0
        # In questo modo i dati trasmessi/ricevuti verranno scritti/letti dall'inizio del Buffer
        Strawberry.Write(REG.FIFO_TX_BASE_ADDR, 0)
        Strawberry.Write(REG.FIFO_RX_BASE_ADDR, 0)

    @staticmethod
    def set_modem_config(modem_config: ModemConfig):
        # Modem Config: si riferisce alla configurazione del modem LoRa, responsabile della modulazione
        # e della demodulazione dei dati da trasmettere e ricevere su una rete LoRa.
        # La configurazione del modem è una serie di parametri:
        #     - Larghezza di banda (BW)
        #     - Fattore di spreading (SF)
        #     - Codice di correzione degli errori (CR)
        #     - Frequenza di trasmissione:
        #     - Potenza di trasmissione
        #     - Lunghezza del pacchetto
        #     - Tipo di modulazione
        Strawberry.Write(REG.MODEM_CONFIG_1, modem_config[0])
        Strawberry.Write(REG.MODEM_CONFIG_2, modem_config[1])
        Strawberry.Write(REG.MODEM_CONFIG_3, modem_config[2])

    @staticmethod
    def set_preamble() -> None:
        # Set Preamble: è una sequenza di bit inviata prima dei dati effettivi nella trasmissione ed è utilizzato per sincronizzare il ricevitore con il trasmettitore.
        # La configurazione del preambolo è importante perché aiuta il ricevitore a riconoscere l'inizio di una trasmissione e sincronizzarsi con il trasmettitore.
        Strawberry.Write(REG.PREAMBLE_MSB, 0)
        Strawberry.Write(REG.PREAMBLE_LSB, 8)

    @staticmethod
    def set_mode(mode: int):
        if LoRa._current_mode == mode:
            return

        # Aggiorno il registro della modalità
        LoRa._current_mode = mode
        Strawberry.Write(REG.OP_MODE, mode)

        # 1. Interrupt on RXDone: Se trasmetto imposto il dio0 per notificarmi dell'invio dei dati
        # 2. Interrupt on TxDone: Se ricevo imposto il dio0 per notificarmi che ci sta un nuovo messaggio da leggere
        Strawberry.Write(REG.DIO_MAPPING_1, 0x40 if mode == MODE.TX else 0x00)

    @staticmethod
    def set_frequency(frequency: float):
        # Set Frequency: a cui il dispositivo trasmette e riceve dati.
        LoRa._frequency = frequency

        i = int(frequency * 16384.0)
        msb = i // 65536
        i -= msb * 65536
        mid = i // 256
        i -= mid * 256
        lsb = i
        Strawberry.Write(REG.FR_MSB, [msb, mid, lsb])

    @staticmethod
    def set_tx_power(tx_power: int):
        # TX Power (Potenza di Trasmissione): rappresenta la quantità di potenza RF (radiofrequenza) utilizzata dal dispositivo
        # per inviare un segnale, espressa in dBm (decibel-milliwatt)
        #     - Potenza massima 23 dBm
        #     - Potenza minima   5 dBm
        tx_power = max(min(tx_power, 23), 5)

        # Set TX Power:
        #     - in base alla potenza che si vorrà utilizzare per trasmettere i dati, si attiverà/disattiverà un amplificatore di potenza
        #     - imposto la potenza di trasmissione
        if tx_power < 20:
            Strawberry.Write(REG.PA_DAC, PA_DAC.ENABLE)
            tx_power -= 3
        else:
            Strawberry.Write(REG.PA_DAC, PA_DAC.DISABLE)

        Strawberry.Write(REG.PA_CONFIG, PA_DAC.SELECT | (tx_power - 5))

    # -------------------------------------------------------------------------------------------- #
    # --- Writer --- #
    @staticmethod
    def send_packet(to: str, data: str):
        # Encode del DevID del destinatario
        devID = LoRa.encode_devID(to)

        # Creo il Payload e lo Trasmetto
        payload = (
            [p for p in devID]
            + [p for p in LoRa._device_id]
            + [p for p in data.encode()]
        )
        LoRa.send(payload)

    @staticmethod
    def send(payload: list[int]):
        # Mi salvo la modalità, in questo modo appena ricevo il TxDone la rimetto
        LoRa._initial_mode = LoRa._current_mode

        # Size del messaggio
        payload_size = len(payload)
        Strawberry.Write(REG.PAYLOAD_LENGTH, payload_size)

        # Entro in modalità Standby, in questo modo mentre modifico il FIFO non ci sarà il rischio che il modulo si metta in mezzo
        LoRa.set_mode(MODE.STANDBY)
        base_addr = Strawberry.Read(REG.FIFO_TX_BASE_ADDR)
        Strawberry.Write(REG.FIFO_ADDR_PTR, base_addr)
        Strawberry.Write(REG.FIFO, payload)

        # Quando c'è il passaggio dalla modalità di Standby alla modalità di TX (Trasmissione), il modulo trasmette i dati.
        # Infatti esso andrà a leggere dal fifo quello che abbiamo appena inserito
        LoRa.set_mode(MODE.TX)

    # -------------------------------------------------------------------------------------------- #
    # --- Interrupts --- #
    # FIFO (First-In, First-Out): è un buffer dove vengono memorizzati temporaneamente i dati che devono essere
    # trasmetti o che sono stati ricevuti, in ordine FIFO.
    #
    # Questo buffer è diviso in due parti:
    #   - FIFO RX: per i dati che vengono ricevuti
    #   - FIFO TX: per i dati che vengono trasmessi
    #
    # Esso ha 3 puntatori, dove 1 è gestito da noi (FifoAddrPtr) e gli altri due sono gestiti dal modulo (RX Addr e TX Addr):
    #   - FifoAddrPtr (Fifo Address Pointer): punta nella locazione in cui vogliamo effettuare un'operazione di lettura/scrittura.
    #     Ogni volta che vorremmo leggere/scrivere nel FIFO dobbiamo puntare in quella locazione utilizzando questo registro.
    #
    #   - FIFO RX Addr: punta nella locazione di memoria dove sono stati appena ricevuti dei dati
    #   - FIFO TX Addr: punta nella locazione di memoria dove sono stati appena trasmessi dei dati
    #
    # Quando il puntatore raggiunge la fine della FIFO, spesso si "ricicla" all'inizio in modo circolare,
    # poiché la FIFO è una struttura dati circolare.
    #

    # DIO0 00: RxDone
    # DIO0 01: TxDone
    @staticmethod
    def _handle_interrupt_dio0(_: int):
        irq_flags = LoRa.read_irq_flags()

        # DIO0 00: RxDone
        if LoRa._current_mode == MODE.RX_CONTINUOUS and (irq_flags & IRQ.RxDone):
            LoRa.clear_irq_flags(IRQ.RxDone)  # Clear RxDone Flag
            LoRa._on_rx_done()

        # DIO0 01: TxDone
        elif LoRa._current_mode == MODE.TX and (irq_flags & IRQ.TxDone):
            LoRa.clear_irq_flags(IRQ.TxDone)  # Clear TxDone Flag
            LoRa._on_tx_done()

    @staticmethod
    def _handle_interrupt_dioX(channel: int):
        print("Handle Interrupt:", channel)

    # -------------------------------------------------------------------------------------------- #
    # --- On ---- #

    # Messaggio:
    #   [0..7] to -> Destinatario
    #   [8..15] from -> Mittente
    #   [15:] payload -> Messaggio

    @staticmethod
    def _on_rx_done():
        # Recupero la lunghezza del packet appena ricevuto
        packet_len = Strawberry.Read(REG.RX_NB_BYTES)

        # 1. Ottengo l'indirizzo del messaggio appena ricevuto
        # 2. Imposto il puntatore del FIFO a questo indirizzo, così potrò leggere
        fifo_rx_current_addr = Strawberry.Read(REG.FIFO_RX_CURR_ADDR)
        Strawberry.Write(
            REG.FIFO_ADDR_PTR, fifo_rx_current_addr
        )  # Recupero l'indirizzo dell'inizio packet

        # Recupero il packet dal FIFO
        packet = Strawberry.Reads(REG.FIFO, packet_len)

        # (Rapporto segnale-rumore, Potenza del segnale ricevuto)
        (snr, rssi) = LoRa._get_signal_info()

        # Unpack ?¿?
        if LoRa._unpack:
            h_to = bytes(packet[0:7])  # Destinatario
            h_from = bytes(packet[8:15])  # Mittente
            payload = bytes(packet[15:])  # Messaggio

            LoRa.on_receive_unpacked(h_to, h_from, payload, snr, rssi)
        else:
            LoRa.on_receive(packet, snr, rssi)

    @staticmethod
    def _on_tx_done():
        # Una volta che ho ricevuto la conferma che il packed che ho appena inserito nel FIFO è stato trasmetto, ritorno alla modalità iniziale (Prima del send)
        LoRa.set_mode(
            LoRa._initial_mode if LoRa._initial_mode is not None else MODE.SLEEP
        )
        LoRa._initial_mode = None

    # -------------------------------------------------------------------------------------------- #
    # --- Utils --- #
    @staticmethod
    def clear_irq_flags(flags: int):
        Strawberry.Write(REG.IRQ_FLAGS, flags)

    @staticmethod
    def read_irq_flags() -> int:
        return Strawberry.Read(REG.IRQ_FLAGS)

    @staticmethod
    def get_all_registers() -> list[int]:
        return [0] + Strawberry.Reads(0x00, 0x3E)[1:]

    @staticmethod
    def _get_signal_info() -> tuple[float, float]:
        # Rapporto segnale-rumore (SNR) espresso in db. Un valore SNR più alto indica una migliore qualità del segnale rispetto al rumore
        snr = Strawberry.Read(REG.PKT_SNR_VALUE) / 4
        # Potenza del segnale ricevuto (RSSI) espresso in dBm. Rappresenta la potenza del segnale ricevuto
        rssi = Strawberry.Read(REG.PKT_RSSI_VALUE)

        # Valori calcoli per correggere l'rssi a seconda di varie condizioni
        if snr < 0:
            rssi = snr + rssi
        else:
            rssi = rssi * 16 / 15
        if LoRa._frequency >= 779:
            rssi = round(rssi - 157, 2)
        else:
            rssi = round(rssi - 164, 2)

        return (snr, rssi)

    @staticmethod
    def encode_devID(devID: str) -> bytes:
        # Controllo la lunghezza
        if len(devID) > 8:
            raise ValueError(
                "Il Device ID deve avere una lunghezza compresa tra 0 e 8 caratteri."
            )

        try:
            # Converto il Device ID in bytes con padding a 8
            return devID.encode("ascii") + (bytes([0x00]) * (8 - len(devID)))
        except UnicodeEncodeError:
            raise UnicodeError("Il Device ID può contenere solo caratteri ASCII.")

==> Strawberry-PI-0.0.66.tar.gz: Strawberry-PI-0.0.66/src/LoRa/LoRa.py <==
import time

from .boards import Board
from .boards.board import Board as IBoard
from .constants import DEFAULT, PA_DAC, REG
from .my_types import IRQ, MODE, ModemConfig


class LoRa(object):
    # SPI
    _board: IBoard

    # ID del dispositivo
    _device_id: list[int]

    # Frequency of Transmission
    _frequency: float

    # Mode:
    #   - É la modalità di inizio, utilizzata per salvare la modalità quando si effettuano più operazioni
    #   -  É la modalità CURRENT
    _initial_mode: int | None = None
    _current_mode: int | None = None

    # To Unpack
    _unpack: bool = DEFAULT.UNPACK

    # -------------------------------------------------------------------------------------------- #
    @staticmethod
    def Setup(
        device_id: str,
        frequency: float = DEFAULT.FREQUENCY,
        tx_power: int = DEFAULT.TX_POWER,
        modem_config: ModemConfig = DEFAULT.MODEM_CONFIG,
        handler: object = None,
        unpack: bool = DEFAULT.UNPACK,
    ):
        # --- Device ID ---- #
        LoRa._device_id = LoRa.encode_devID(device_id)

        # --- Board--- #
        LoRa._board = Board()
        LoRa._board.spi()
        LoRa._board.add_events(
            callback_dio0=LoRa._handle_interrupt_dio0,
            callback_dio1=LoRa._handle_interrupt_dioX,
            callback_dio2=LoRa._handle_interrupt_dioX,
            callback_dio3=LoRa._handle_interrupt_dioX,
        )  # Add Events

        # --- Settings --- #
        LoRa.check_module()  # Check if module LoRa is connected
        LoRa.set_mode(MODE.SLEEP)  # For calibration
        # LoRa.reset_fifo()  # Reset Fifo buffer
        LoRa.set_modem_config(modem_config)  # Set Modem Config
        # self.set_preamble()  # Set Preamble
        LoRa.set_frequency(frequency)  # Frequency
        LoRa.set_tx_power(tx_power)  # Transmission Power
        LoRa.set_handler(handler, unpack)
        LoRa.set_mode(
            MODE.STANDBY
        )  # Module in Standby -> In questo modo sarà il Main a scegliere il da farsi

    # -------------------------------------------------------------------------------------------- #
    # --- Events --- #
    @staticmethod
    def on_receive(payload: bytes, snr: float, rssi: float) -> None:
        raise NotImplementedError(
            "Coglione ti sei dimenticato di implementare la funzione!"
        )

    @staticmethod
    def on_receive_unpacked(
        h_to: bytes, h_from: bytes, payload: bytes, snr: float, rssi: float
    ) -> None:
        raise NotImplementedError(
            "Coglione ti sei dimenticato di implementare la funzione!"
        )

    # -------------------------------------------------------------------------------------------- #
    # --- Module Function --- #
    @staticmethod
    def set_handler(callback: object, unpack: bool = DEFAULT.UNPACK):
        if callback is not None:
            # Imposto la callback per i messaggi che ricevo
            LoRa._unpack = unpack

            if unpack:
                LoRa.on_receive_unpacked = callback  # type: ignore
            else:
                LoRa.on_receive = callback  # type: ignore

    @staticmethod
    def check_module():
        # Controllo se il modulo è stato correttamente collegato al microcontrollore
        #     Scrivo nel registro e quando rileggo mi devo aspettare lo stesso risultato
        LoRa._board.write(REG.OP_MODE, MODE.SLEEP)
        time.sleep(0.1)

        if LoRa._board.read(REG.OP_MODE) != MODE.SLEEP:
            raise ConnectionError("LoRa initialization failed!")

    @staticmethod
    def reset_fifo():
        # Imposto la posizione di base del buffer FIFO di trasmissione ('FIFO TX') e di ricezione ('FIFO RX) a 0
        # In questo modo i dati trasmessi/ricevuti verranno scritti/letti dall'inizio del Buffer
        LoRa._board.write(REG.FIFO_TX_BASE_ADDR, 0)
        LoRa._board.write(REG.FIFO_RX_BASE_ADDR, 0)

    @staticmethod
    def set_modem_config(modem_config: ModemConfig):
        # Modem Config: si riferisce alla configurazione del modem LoRa, responsabile della modulazione
        # e della demodulazione dei dati da trasmettere e ricevere su una rete LoRa.
        # La configurazione del modem è una serie di parametri:
        #     - Larghezza di banda (BW)
        #     - Fattore di spreading (SF)
        #     - Codice di correzione degli errori (CR)
        #     - Frequenza di trasmissione:
        #     - Potenza di trasmissione
        #     - Lunghezza del pacchetto
        #     - Tipo di modulazione
        LoRa._board.write(REG.MODEM_CONFIG_1, modem_config[0])
        LoRa._board.write(REG.MODEM_CONFIG_2, modem_config[1])
        LoRa._board.write(REG.MODEM_CONFIG_3, modem_config[2])

    @staticmethod
    def set_preamble() -> None:
        # Set Preamble: è una sequenza di bit inviata prima dei dati effettivi nella trasmissione ed è utilizzato per sincronizzare il ricevitore con il trasmettitore.
        # La configurazione del preambolo è importante perché aiuta il ricevitore a riconoscere l'inizio di una trasmissione e sincronizzarsi con il trasmettitore.
        LoRa._board.write(REG.PREAMBLE_MSB, 0)
        LoRa._board.write(REG.PREAMBLE_LSB, 8)

    @staticmethod
    def set_mode(mode: int):
        if LoRa._current_mode == mode:
            return

        # Aggiorno il registro della modalità
        LoRa._current_mode = mode
        LoRa._board.write(REG.OP_MODE, mode)

        # 1. Interrupt on RXDone: Se trasmetto imposto il dio0 per notificarmi dell'invio dei dati
        # 2. Interrupt on TxDone: Se ricevo imposto il dio0 per notificarmi che ci sta un nuovo messaggio da leggere
        LoRa._board.write(REG.DIO_MAPPING_1, 0x40 if mode == MODE.TX else 0x00)

    @staticmethod
    def set_frequency(frequency: float):
        # Set Frequency: a cui il dispositivo trasmette e riceve dati.
        LoRa._frequency = frequency

        i = int(frequency * 16384.0)
        msb = i // 65536
        i -= msb * 65536
        mid = i // 256
        i -= mid * 256
        lsb = i
        LoRa._board.write(REG.FR_MSB, [msb, mid, lsb])

    @staticmethod
    def set_tx_power(tx_power: int):
        # TX Power (Potenza di Trasmissione): rappresenta la quantità di potenza RF (radiofrequenza) utilizzata dal dispositivo
        # per inviare un segnale, espressa in dBm (decibel-milliwatt)
        #     - Potenza massima 23 dBm
        #     - Potenza minima   5 dBm
        tx_power = max(min(tx_power, 23), 5)

        # Set TX Power:
        #     - in base alla potenza che si vorrà utilizzare per trasmettere i dati, si attiverà/disattiverà un amplificatore di potenza
        #     - imposto la potenza di trasmissione
        if tx_power < 20:
            LoRa._board.write(REG.PA_DAC, PA_DAC.ENABLE)
            tx_power -= 3
        else:
            LoRa._board.write(REG.PA_DAC, PA_DAC.DISABLE)

        LoRa._board.write(REG.PA_CONFIG, PA_DAC.SELECT | (tx_power - 5))

    # -------------------------------------------------------------------------------------------- #
    # --- Writer --- #
    @staticmethod
    def send_packet(to: str, data: str):
        # Encode del DevID del destinatario
        devID = LoRa.encode_devID(to)

        # Creo il Payload e lo Trasmetto
        payload = devID + LoRa._device_id + [p for p in data.encode()]
        LoRa.send(payload)

    @staticmethod
    def send(payload: list[int]):
        # Mi salvo la modalità, in questo modo appena ricevo il TxDone la rimetto
        LoRa._initial_mode = LoRa._current_mode

        # Size del messaggio
        payload_size = len(payload)
        LoRa._board.write(REG.PAYLOAD_LENGTH, payload_size)

        # Entro in modalità Standby, in questo modo mentre modifico il FIFO non ci sarà il rischio che il modulo si metta in mezzo
        LoRa.set_mode(MODE.STANDBY)
        base_addr = LoRa._board.read(REG.FIFO_TX_BASE_ADDR)
        LoRa._board.write(REG.FIFO_ADDR_PTR, base_addr)
        LoRa._board.write(REG.FIFO, payload)

        # Quando c'è il passaggio dalla modalità di Standby alla modalità di TX (Trasmissione), il modulo trasmette i dati.
        # Infatti esso andrà a leggere dal fifo quello che abbiamo appena inserito
        LoRa.set_mode(MODE.TX)

    # -------------------------------------------------------------------------------------------- #
    # --- Interrupts --- #
    # FIFO (First-In, First-Out): è un buffer dove vengono memorizzati temporaneamente i dati che devono essere
    # trasmetti o che sono stati ricevuti, in ordine FIFO.
    #
    # Questo buffer è diviso in due parti:
    #   - FIFO RX: per i dati che vengono ricevuti
    #   - FIFO TX: per i dati che vengono trasmessi
    #
    # Esso ha 3 puntatori, dove 1 è gestito da noi (FifoAddrPtr) e gli altri due sono gestiti dal modulo (RX Addr e TX Addr):
    #   - FifoAddrPtr (Fifo Address Pointer): punta nella locazione in cui vogliamo effettuare un'operazione di lettura/scrittura.
    #     Ogni volta che vorremmo leggere/scrivere nel FIFO dobbiamo puntare in quella locazione utilizzando questo registro.
    #
    #   - FIFO RX Addr: punta nella locazione di memoria dove sono stati appena ricevuti dei dati
    #   - FIFO TX Addr: punta nella locazione di memoria dove sono stati appena trasmessi dei dati
    #
    # Quando il puntatore raggiunge la fine della FIFO, spesso si "ricicla" all'inizio in modo circolare,
    # poiché la FIFO è una struttura dati circolare.
    #

    # DIO0 00: RxDone
    # DIO0 01: TxDone
    @staticmethod
    def _handle_interrupt_dio0(_: int):
        irq_flags = LoRa.read_irq_flags()

        # DIO0 00: RxDone
        if LoRa._current_mode == MODE.RX_CONTINUOUS and (irq_flags & IRQ.RxDone):
            LoRa.clear_irq_flags(IRQ.RxDone)  # Clear RxDone Flag
            LoRa._on_rx_done()

        # DIO0 01: TxDone
        elif LoRa._current_mode == MODE.TX and (irq_flags & IRQ.TxDone):
            LoRa.clear_irq_flags(IRQ.TxDone)  # Clear TxDone Flag
            LoRa._on_tx_done()

    @staticmethod
    def _handle_interrupt_dioX(channel: int):
        print("Handle Interrupt:", channel)

    # -------------------------------------------------------------------------------------------- #
    # --- On ---- #

    # Messaggio:
    #   [0..7] to -> Destinatario
    #   [8..15] from -> Mittente
    #   [15:] payload -> Messaggio

    @staticmethod
    def _on_rx_done():
        # Recupero la lunghezza del packet appena ricevuto
        packet_len = LoRa._board.read(REG.RX_NB_BYTES)

        # 1. Ottengo l'indirizzo del messaggio appena ricevuto
        # 2. Imposto il puntatore del FIFO a questo indirizzo, così potrò leggere
        fifo_rx_current_addr = LoRa._board.read(REG.FIFO_RX_CURR_ADDR)
        LoRa._board.write(
            REG.FIFO_ADDR_PTR, fifo_rx_current_addr
        )  # Recupero l'indirizzo dell'inizio packet

        # Recupero il packet dal FIFO
        packet = LoRa._board.reads(REG.FIFO, packet_len)

        # (Rapporto segnale-rumore, Potenza del segnale ricevuto)
        (snr, rssi) = LoRa._get_signal_info()

        # Unpack ?¿?
        if LoRa._unpack:
            h_to = bytes(packet[0:7])  # Destinatario
            h_from = bytes(packet[8:15])  # Mittente
            payload = bytes(packet[15:])  # Payload

            LoRa.on_receive_unpacked(h_to, h_from, payload, snr, rssi)
        else:
            payload = bytes(packet)  # Messaggio

            LoRa.on_receive(payload, snr, rssi)

    @staticmethod
    def _on_tx_done():
        # Una volta che ho ricevuto la conferma che il packed che ho appena inserito nel FIFO è stato trasmetto, ritorno alla modalità iniziale (Prima del send)
        LoRa.set_mode(
            LoRa._initial_mode if LoRa._initial_mode is not None else MODE.SLEEP
        )
        LoRa._initial_mode = None

    # -------------------------------------------------------------------------------------------- #
    # --- Utils --- #
    @staticmethod
    def clear_irq_flags(flags: int):
        LoRa._board.write(REG.IRQ_FLAGS, flags)

    @staticmethod
    def read_irq_flags() -> int:
        return LoRa._board.read(REG.IRQ_FLAGS)

    @staticmethod
    def get_all_registers() -> list[int]:
        return [0] + LoRa._board.reads(0x00, 0x3E)[1:]

    @staticmethod
    def _get_signal_info() -> tuple[float, float]:
        # Rapporto segnale-rumore (SNR) espresso in db. Un valore SNR più alto indica una migliore qualità del segnale rispetto al rumore
        snr = LoRa._board.read(REG.PKT_SNR_VALUE) / 4
        # Potenza del segnale ricevuto (RSSI) espresso in dBm. Rappresenta la potenza del segnale ricevuto
        rssi = LoRa._board.read(REG.PKT_RSSI_VALUE)

        # Valori calcoli per correggere l'rssi a seconda di varie condizioni
        if snr < 0:
            rssi = snr + rssi
        else:
            rssi = rssi * 16 / 15
        if LoRa._frequency >= 779:
            rssi = round(rssi - 157, 2)
        else:
            rssi = round(rssi - 164, 2)

        return (snr, rssi)

    @staticmethod
    def encode_devID(devID: str) -> list[int]:
        # Controllo la lunghezza
        if len(devID) > 8:
            raise ValueError(
                "Il Device ID deve avere una lunghezza compresa tra 0 e 8 caratteri."
            )

        try:
            # Converto il Device ID in bytes con padding a 8
            l = [x for x in devID.encode("ascii")]
            return l + ([0x00] * (8 - len(devID)))
        except UnicodeEncodeError:
            raise UnicodeError("Il Device ID può contenere solo caratteri ASCII.")
